package com.yuandian.dataflow.statemachine.operate;

import java.io.Serializable;

import com.alipay.sofa.jraft.Status;
import com.alipay.sofa.jraft.error.RaftError;
import com.alipay.sofa.jraft.error.RaftException;
import com.alipay.sofa.jraft.error.RemotingException;
import com.alipay.sofa.jraft.rpc.InvokeCallback;
import com.yuandian.dataflow.statemachine.StateFactory;
import com.yuandian.dataflow.statemachine.closure.GenericClosure;
import com.yuandian.dataflow.statemachine.rpc.OperateProcessor;
import com.yuandian.dataflow.statemachine.rpc.RaftResponse;
import com.yuandian.dataflow.statemachine.state.WorkerState;

import lombok.Data;

import lombok.extern.slf4j.Slf4j;

/**
 * 操作
 * 
 * @author eson
 */
@Slf4j
@Data
public class Operate implements Serializable {

    public static enum OperateType {
        /**
         * 同步WorkerState状态. 
         */
        PUT_WORKERSTATE, 
        /**
         *  暂无想法
         */
        REMOVE;
    }

    private OperateType type;
    private Object value;

    public Operate(OperateType t, WorkerState ws) {
        this.type = t;
        this.value = ws;
    }

    public <T> T getValue() {
        return (T) this.value;
    };

    public <T> void setValue(T value) {
        this.value = value;
        return;
    };

    /**
     * 调用操作设置
     * @param op 传入的操作类
     * @param closure 回调函数. Operate为返回值
     */
    public static void CallOperate(Operate op, GenericClosure<Operate> closure) {
        log.debug("CallOperate Value {}", op.<WorkerState>getValue());
        var ss = StateFactory.getStateServer();
        // 如果是leader 就直接提交
        if (StateFactory.isLeader()) { 
            ss.applyOperate(op, closure);
            return;
        }

        // 非leader 转发请求 统一有leader处理
        var request = new OperateProcessor.OperateRequest();
        request.setOperate(op);

        var leaderId = StateFactory.getLeaderId();
        try {
            ss.getRpcClient().invokeAsync(leaderId.getEndpoint(),
                    request, new InvokeCallback() {

                        @Override
                        public void complete(Object result, Throwable err) {
                            log.debug("Object result {}", result);
                      
                            var resp = (RaftResponse<Operate>) result;
                            closure.setResponse(resp);
                            closure.success(resp.getValue());
                            closure.run(Status.OK());
                        }

                    }, 5000);
        } catch (InterruptedException | RemotingException e) {
            closure.failure(e.getMessage(), null);
            closure.run(new Status(100000, "invokeAsync fail"));
            log.info("{}", e.toString());
        }

    }
}
